import ipfshttpclient
import json
from json.decoder import JSONDecodeError
from medblocks import cryptography
from medblocks.errors import *
import requests
from concurrent.futures import ThreadPoolExecutor, as_completed
import time

IPFS_PORT = "5001"
MEDBLOCKS_PORT = "8080"

    
class Client(object):
    def __init__(self, ip_address, *args, **kwargs):
        self.reference_mapping = {}
        self.base_url = "http://{ip_address}:{port}/".format(ip_address=ip_address, port=MEDBLOCKS_PORT)
        self.ipfs = ipfshttpclient.connect("/ip4/{ip_address}/tcp/{port}/http/".format(ip_address=ip_address, port=IPFS_PORT))
        self.logged_in = False
        
    def _check_login(self):
        if self.logged_in:
            return
        else:
            raise LoginError("Please login and then try again")

    def medblocks(self, endpoint, json={}):
        url = self.base_url + "{}".format(endpoint)
        r = requests.post(url, json=json)
        try:
            r.raise_for_status()
            return r.json()
        except requests.exceptions.HTTPError: # TODO Check backend for JSON in errors
            raise MedBlocksAPIError(r.text)

    def _checkReferences(dict):
        """Checks dictionary for references and replaces them with known references or throws error"""

    def list(self, **kwargs):
        """
        Optional kwargs:
        ownerEmailId (string): Lists all medblocks that belong to the owner
        permittedEmailId (string): List all medblocks that this email has permission to
        """
        return self.medblocks("list", kwargs)
    
    def register(self, emailId, password):
        """
        Generates a key pair, encrypts with as password and stores it on the server. Registers a 
        """
        keypairs = cryptography.generateKeys(password)
        keypairs["emailId"] = emailId
        self.medblocks("storeKey", keypairs)
        return self.medblocks("register",{
            "emailId": emailId,
            "sPublicKey": keypairs["sPublicKey"],
            "ePublicKey": keypairs["ePublicKey"]
            })
    
    def login(self, emailId, password, offlineKeys=None):
        if offlineKeys is None:
            keys = self.medblocks("getKey", {"emailId": emailId})
        else:
            keys = offlineKeys
        self.ePrivateKey = cryptography.decryptKey(keys["ePrivateKey"].encode(), keys["IV"]["IVE"].encode(), password)
        self.sPrivateKey = cryptography.decryptKey(keys["sPrivateKey"].encode(), keys["IV"]["IVS"].encode(), password)
        self.ePublicKey = cryptography.get_publicKey(self.ePrivateKey)
        self.sPublicKey = cryptography.get_publicKey(self.sPrivateKey)
        self.emailId = emailId
        self.logged_in = True
        return
    
    def getPublicKey(self, email):
        user = self.medblocks("getUser", {"emailId": email})
        return user["ePublicKey"]

    def prepare_payload(self, dict):
        data = json.dumps(dict)
        signature = cryptography.generate_signature(data, self.sPrivateKey)
        payload = {"data":data, "signature": signature.decode()}
        return payload

    def addBlock(self, bytes, to_address, name, open=False):
        self._check_login()

        if open:
            r = self.ipfs.add_bytes(bytes)
            data = {
            "ipfsHash": r,
            "IV": "NO ENCRYPTION",
            "name": name,
            "format": "fhir/json",
            "creatorEmailId": self.emailId,
            "ownerEmailId": self.emailId,
            "permissions":[
            ]
        }
        else:
            encryption_start = time.time()
            public_key = self.getPublicKey(to_address)
            aes_key = cryptography.generate_aes_key()
            encrypted_file, iv = cryptography.encrypt_file(bytes, aes_key)
            access_key = cryptography.create_access_key(aes_key, public_key)   
            encryption_elapsed = round(time.time() - encryption_start, 2)
            ipfs_start = time.time()
            r = self.ipfs.add_bytes(encrypted_file)
            ipfs_elapsed = round(time.time() - ipfs_start, 2)
            iv = iv.decode()
            data = {
                "ipfsHash": r,
                "IV": iv,
                "name": name,
                "format": "fhir/json",
                "creatorEmailId": self.emailId,
                "ownerEmailId": self.emailId,
                "permissions":[
                {"receiverEmailId": to_address, 
                "receiverKey":access_key.decode()}
                ]
            }
        
        signing_start = time.time()
        payload = self.prepare_payload(data)
        signing_elapsed = round(time.time() - signing_start)
        request_start = time.time()
        medblock = self.medblocks("addBlock", payload)
        request_elapsed = round(time.time() - request_start, 2)
        # print("Time elapsed: encryption: {}sec | ipfs: {}sec | signing: {}sec | request: {}".format(encryption_elapsed, ipfs_elapsed, signing_elapsed, request_elapsed))
        return medblock
    
    def extractKey(self, block):
        self._check_login()
        permissions = block["permissions"]
        for p in permissions:
            if p["receiverEmailId"] == self.emailId:
                accessKey = p["receiverKey"]
        try:
            aes_key = cryptography.decrypt_access_key(accessKey, self.ePrivateKey)
            return aes_key
        except NameError:
            raise PermissionError("No access key for current user found. To view raw block set raw=True")

    def getBlock(self, ipfsHash, raw=False):
        block = self.medblocks("getBlock",{"ipfsHash":ipfsHash})
        if raw:
            return block
        aes_key = self.extractKey(block)
        iv = block["IV"].encode()
        encrypted = self.ipfs.cat(ipfsHash)
        decrypted = cryptography.decrypt_file(encrypted, aes_key, iv)
        return decrypted

    def addPermission(self, ipfsHash, to):
        self._check_login()
        public_key = self.getPublicKey(to)
        block = self.medblocks("getBlock",{"ipfsHash":ipfsHash})
        aes_key = self.extractKey(block)
        accessKey = cryptography.create_access_key(aes_key, public_key)
        data = {
            "ipfsHash": ipfsHash,
            "senderEmailId": self.emailId,
            "permissions": [
                {
                    "receiverEmailId": to,
                    "receiverKey": accessKey.decode()
                }
            ]
        }
        payload = self.prepare_payload(data)
        return self.medblocks("addPermission", payload)

    def logout(self):
        self.ePublicKey = None
        self.sPublicKey = None
        self.ePrivateKey = None
        self.sPrivateKey = None
        self.logged_in = False
        return

    def async_add_medblock(self, uuid, data):
        start_time = time.time()
        medblock = self.addBlock(data, self.emailId, "ORIGIN")
        hash = medblock["ipfsHash"]
        elapsed = round(time.time() - start_time, 2)
        self.reference_mapping[uuid] = hash
        print("addBlock completed {} -> {} ({} secs)".format(uuid, hash, elapsed))
    
    def addFhir(self, data, version="R4", max_workers=10):
        start = time.time()
        for resource in data["entry"]:
            fullUrl = resource.pop("fullUrl")
            resource["resource"].pop("id")
            self.reference_mapping[fullUrl] = resource
        # with ThreadPoolExecutor(3) as f:
        #     for i in range(100):
        #         f.submit(self.test_async, i, i)
        
        with ThreadPoolExecutor(max_workers) as f:
            for uuid in self.reference_mapping.keys():
                data = self.reference_mapping[uuid]
                data = json.dumps(data).encode()
                # self.async_add_medblock(uuid, data)
                f.submit(self.async_add_medblock, uuid, data)
        elapsed = round(time.time() - start, 2)
        print("Finished addBlock for {} in {} secs".format(len(self.reference_mapping.keys()), elapsed))
        mapping = self.reference_mapping
        self.addBlock(json.dumps(mapping).encode(), self.emailId, "MAP", open=True)
        self.reference_mapping = {}
        return mapping


from Crypto.PublicKey import RSA
from Crypto.Cipher import AES, PKCS1_OAEP
from Crypto.Signature import pkcs1_15
from Crypto.Hash import SHA256
from Crypto.Protocol.KDF import PBKDF2
from base64 import b64decode, b64encode
import os

def generate_RSA():
    privateKey = RSA.generate(2048)
    return privateKey.publickey().export_key(), privateKey.export_key()

def encryptKey(key, password):
    derivedKey = PBKDF2(password, password, 16)
    cipher = AES.new(derivedKey, AES.MODE_GCM)
    iv = cipher.nonce
    encrypted, tag = cipher.encrypt_and_digest(key)
    return b64encode(encrypted), b64encode(iv)

def decryptKey(encryptedkey, iv, password):
    derivedKey = PBKDF2(password, password, 16)
    iv = b64decode(iv)
    cipher = AES.new(derivedKey, AES.MODE_GCM, nonce=iv)
    decrypted = cipher.decrypt(b64decode(encryptedkey))
    return decrypted

def generateKeys(password):
    sPublicKey, sPrivateKey = generate_RSA()
    ePublicKey, ePrivateKey = generate_RSA()
    sPrivateKey, ivs = encryptKey(sPrivateKey, password)
    ePrivateKey, ive = encryptKey(ePrivateKey, password)
    return {
        "IV": {
            "IVE": ive.decode(),
            "IVS": ivs.decode()
            },
        "ePublicKey": ePublicKey.decode(),
        "ePrivateKey": ePrivateKey.decode(),
        "sPublicKey": sPublicKey.decode(),
        "sPrivateKey": sPrivateKey.decode()
        }
    
def generate_aes_key() -> "bytes" :
    """generate random AES key (-> aes_secret)"""
    secret = os.urandom(32)
    return secret

def create_access_key(aes_key, public_key):
    """encrypt AES key to get access key (aes_secret, public_key -> access_key)"""
    rsaKey = RSA.import_key(public_key)
    cipher = PKCS1_OAEP.new(rsaKey)
    encrypted = cipher.encrypt(aes_key)
    return b64encode(encrypted)

def decrypt_access_key(access_key, private_key):
    """Decrypt access key using RSA private Key(RSA_Key)"""
    access_key = b64decode(access_key)
    rsaKey = RSA.import_key(private_key)
    cipher = PKCS1_OAEP.new(rsaKey)
    decrypted = cipher.decrypt(access_key)
    return decrypted

def encrypt_file(bytes, aes_key):
    """encrypt file with key (f, aes_secret -> enc(f), iv)"""
    cipher = AES.new(aes_key, AES.MODE_GCM)
    iv = cipher.nonce
    encrypted, tag = cipher.encrypt_and_digest(bytes)
    return b64encode(encrypted), b64encode(iv)

def decrypt_file(file, aes_key, iv):
    file = b64decode(file)
    iv = b64decode(iv)
    cipher = AES.new(aes_key, AES.MODE_GCM, nonce=iv)
    decrypted = cipher.decrypt(file)
    return decrypted

def generate_signature(s, private_key):
    """generates signature base64encoded using RSA"""
    rsaKey = RSA.import_key(private_key)
    hash = SHA256.new(s.encode())
    cipher = pkcs1_15.new(rsaKey)
    signature = cipher.sign(hash)
    return b64encode(signature)

def get_publicKey(privateKey):
    rsaKey = RSA.import_key(privateKey)
    publicKey = rsaKey.publickey()
    return publicKey.export_key()